@@ -1,10 +1,12 @@
 use std::collections::HashMap;
 use std::fmt::Display;
+use std::path::PathBuf;
 use std::str::FromStr;
 use std::sync::Arc;
 use std::time::Duration;
 
 use bson::oid::ObjectId;
+use bson::Document;
 use futures::StreamExt;
 use mongodb::options::{
     Acknowledgment, AuthMechanism, Credential, ReadConcern, ReadConcernLevel, ReadPreference,
@@ -12,6 +14,7 @@ use mongodb::options::{
 };
 use url::Url;
 
+use super::TypedCursor;
 use crate::collection::Collection;
 use crate::filter::{AsFilter, Filter};
 use crate::query;
@@ -63,8 +66,8 @@ impl ClientBuilder {
         // and work around that with minimal code duplication
         let url = Url::parse(&uri).map_err(crate::error::builder)?;
         let mut options = mongodb::options::ClientOptions::builder()
-            .hosts(vec![mongodb::options::StreamAddress {
-                hostname: url.host_str().unwrap_or("127.0.0.1").to_owned(),
+            .hosts(vec![mongodb::options::ServerAddress::Tcp {
+                host: url.host_str().unwrap_or("127.0.0.1").to_owned(),
                 port: url.port(),
             }])
             .build();
@@ -148,10 +151,11 @@ impl ClientBuilder {
             // FIXME: Not supported by mongodb...
             //options.wait_queue_multiple = Some(str::parse(&wait_queue_multiple).map_err(crate::error::builder)?);
         }
-        if let Some(wait_queue_timeout) = kv.remove("waitQueueTimeoutMS") {
-            options.wait_queue_timeout = Some(Duration::from_millis(
-                str::parse(&wait_queue_timeout).map_err(crate::error::builder)?,
-            ));
+        if let Some(_wait_queue_timeout) = kv.remove("waitQueueTimeoutMS") {
+            // FIXME: Not supported by mongodb...
+            // options.wait_queue_timeout = Some(Duration::from_millis(
+            //     str::parse(&wait_queue_timeout).map_err(crate::error::builder)?,
+            // ));
         }
         if let Some(_zlib_compression_level) = kv.remove("zlibCompressionLevel") {
             // FIXME: Field is private...
@@ -167,7 +171,7 @@ impl ClientBuilder {
                 write_concern.journal = Some(str::parse(&journal).map_err(crate::error::builder)?);
             }
             if let Some(w) = w {
-                let w = match str::parse::<i32>(&w) {
+                let w = match str::parse::<u32>(&w) {
                     Ok(n) => Acknowledgment::Nodes(n),
                     Err(_) => match w.as_str() {
                         "majority" => Acknowledgment::Majority,
@@ -242,8 +246,10 @@ impl ClientBuilder {
         {
             let mut credentials = Credential::default();
             if self.username.is_some() || url.username() != "" {
-                credentials.username = self.username.or(Some(url.username().to_string()));
-                credentials.password = self.password.or(url.password().map(|p| p.to_string()));
+                credentials.username = self.username.or_else(|| Some(url.username().to_string()));
+                credentials.password = self
+                    .password
+                    .or_else(|| url.password().map(|p| p.to_string()));
             }
             if auth_source.is_some() {
                 credentials.source = auth_source;
@@ -301,10 +307,10 @@ impl ClientBuilder {
                             Some(tls_insecure.parse().map_err(crate::error::builder)?);
                     }
                     if let Some(ca_file_path) = self.ca.or(tls_ca_file) {
-                        options.ca_file_path = Some(ca_file_path);
+                        options.ca_file_path = Some(PathBuf::from(ca_file_path));
                     }
                     if let Some(cert_key_file_path) = self.cert_key.or(tls_certificate_key_file) {
-                        options.cert_key_file_path = Some(cert_key_file_path);
+                        options.cert_key_file_path = Some(PathBuf::from(cert_key_file_path));
                     }
                     Tls::Enabled(options)
                 }
@@ -472,7 +478,7 @@ impl Client {
     }
 
     /// Returns the `mongodb::Document` from the mongodb.
-    pub fn collection<C>(&self) -> mongodb::Collection
+    pub fn collection<C>(&self) -> mongodb::Collection<Document>
     where
         C: Collection,
     {
@@ -492,7 +498,7 @@ impl Client {
     /// # Errors
     ///
     /// This method fails if the mongodb encountered an error.
-    pub async fn delete<C, F>(&self, filter: Option<F>) -> crate::Result<i64>
+    pub async fn delete<C, F>(&self, filter: Option<F>) -> crate::Result<u64>
     where
         C: AsFilter<F> + Collection,
         F: Filter,
@@ -501,7 +507,7 @@ impl Client {
         if let Some(filter) = filter {
             delete = delete.filter::<F>(filter)?
         }
-        delete.query(&self).await
+        delete.query(self).await
     }
 
     /// Convenience method to delete one document from a collection using a given filter.
@@ -517,7 +523,7 @@ impl Client {
         let deleted = query::Delete::<C>::new()
             .many(false)
             .filter::<F>(filter)?
-            .query(&self)
+            .query(self)
             .await?;
         Ok(deleted > 0)
     }
@@ -530,7 +536,7 @@ impl Client {
     /// # Errors
     ///
     /// This method fails if the mongodb encountered an error.
-    pub async fn find<C, F>(&self, filter: Option<F>) -> crate::Result<mongodb::Cursor>
+    pub async fn find<C, F>(&self, filter: Option<F>) -> crate::Result<TypedCursor<C>>
     where
         C: AsFilter<F> + Collection,
         F: Filter,
@@ -539,7 +545,7 @@ impl Client {
         if let Some(filter) = filter {
             find = find.filter(filter)?;
         }
-        find.query(&self).await
+        find.query(self).await
     }
 
     /// Convenience method to find a document in a collection using a given filter.
@@ -550,18 +556,16 @@ impl Client {
     /// # Errors
     ///
     /// This method fails if the mongodb encountered an error, or if the found document is invalid.
-    pub async fn find_one<C, F>(&self, filter: F) -> crate::Result<Option<C>>
+    pub async fn find_one<C, F>(&self, filter: F) -> crate::Result<Option<(ObjectId, C)>>
     where
         C: AsFilter<F> + Collection,
         F: Filter,
     {
         // NOTE: We don't wanna make another builder so we just eat the cost of getting a cursor...
         let find: query::Find<C> = query::Find::new();
-        let mut cursor = find.filter(filter)?.query(&self).await?;
+        let mut cursor = find.filter(filter)?.query(self).await?;
         if let Some(res) = cursor.next().await {
-            let doc = res.map_err(crate::error::mongodb)?;
-            let document: C = C::from_document(doc).map_err(crate::error::bson)?;
-            return Ok(Some(document));
+            return Ok(Some(res?));
         }
         Ok(None)
     }
@@ -575,7 +579,7 @@ impl Client {
     where
         C: Collection,
     {
-        let result = query::Insert::new().query(&self, documents).await?;
+        let result = query::Insert::new().query(self, documents).await?;
         Ok(result
             .into_iter()
             .filter_map(|(k, v)| match v {
@@ -595,11 +599,9 @@ impl Client {
         C: Collection,
     {
         // NOTE: We don't wanna make another builder so we just eat the cost of allocating a vec...
-        let result = query::Insert::new().query(&self, vec![document]).await?;
-        if let Some((_, v)) = result.into_iter().next() {
-            if let bson::Bson::ObjectId(id) = v {
-                return Ok(id);
-            }
+        let result = query::Insert::new().query(self, vec![document]).await?;
+        if let Some((_, bson::Bson::ObjectId(id))) = result.into_iter().next() {
+            return Ok(id);
         }
         Err(crate::error::mongodb(
             "failed to insert document into mongo",
@@ -618,7 +620,7 @@ impl Client {
     {
         query::Replace::new()
             .filter::<F>(filter)?
-            .query(&self, document)
+            .query(self, document)
             .await
     }
 
@@ -635,7 +637,7 @@ impl Client {
     {
         let updated = query::Update::<C>::new()
             .filter::<F>(filter)?
-            .query::<U>(&self, updates)
+            .query::<U>(self, updates)
             .await?;
         Ok(updated)
     }
@@ -654,7 +656,7 @@ impl Client {
         let updated = query::Update::<C>::new()
             .many(false)
             .filter::<F>(filter)?
-            .query::<U>(&self, updates)
+            .query::<U>(self, updates)
             .await?;
         if updated > 0 {
             return Ok(true);
@@ -676,7 +678,7 @@ impl Client {
         let updated = query::Update::<C>::new()
             .upsert(true)
             .filter::<F>(filter)?
-            .query::<U>(&self, updates)
+            .query::<U>(self, updates)
             .await?;
         Ok(updated)
     }
@@ -696,7 +698,7 @@ impl Client {
             .many(false)
             .upsert(true)
             .filter::<F>(filter)?
-            .query::<U>(&self, updates)
+            .query::<U>(self, updates)
             .await?;
         if updated > 0 {
             return Ok(true);
